# -*- coding:utf-8 -*-
##############################################################
# Created Date: Tuesday, May 30th 2023
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

# Please note,
# All countries, states, and cities in global_rel_id.json file are from https://github.com/dr5hn/countries-states-cities-database/tree/master
from .extract_relation_id import getOSMRelationID

__all__ = ['getOSMRelationID']